#!/usr/bin/env python
"""\
 DFunctionals.py Exchange and Correlation Density Functionals

 This program is part of the PyQuante quantum chemistry program suite.

 Copyright (c) 2004, Richard P. Muller. All Rights Reserved. 

 PyQuante version 1.2 and later is covered by the modified BSD
 license. Please see the file LICENSE that is part of this
 distribution.

"""

from math import *
from Numeric import *

def XC(dens,gamma,**opts):
    """\
    Simplified top level routine for all XC functionals.
    """
    functional = opts.get('functional','SVWN')
    assert functional in xfuncs.keys() and functional in cfuncs.keys()
    exc,vxc = xfuncs[functional](dens,gamma)
    if cfuncs[functional]:
        ec,vc = cfuncs[functional](dens,gamma)
        vxc = vxc + vc
        exc = exc + ec
    return exc,vxc

def S(dens,gamma=None):
    """
    Slater exchange functional for a vector of densities. Based upon
    the classic Slater exchange (which actually came from Dirac). See
    JC Slater 'The self consistent field for molecules and solids',
    McGraw Hill, New York, 1974.

    Note that gamma is ignored, and is only included for consistency
    with the other XC functionals.
    """
    npts = len(dens)
    ex = zeros(npts,'d')
    vx = zeros(npts,'d')
    for i in range(npts):
        rho = 0.5*dens[i] # Density of the alpha spin
        exa,vxa = xs(rho)
        ex[i] = 2*exa
        vx[i] = vxa
    return ex,vx

def VWN(dens,gamma=None):
    """
    Vosko-Wilk-Nusair correlation functional for a vector of
    densities. From 'Accurate spin-dependent electron liquid
    correlation energies for local spin density calculations:
    a critical analysis.' SH Vosko, L Wilk, M Nusair, Can J
    Phys, 58, 1200 (1980).

    Note that gamma is ignored, and is only included for consistency
    with the other XC functionals.
    """
    npts = len(dens)
    ec = zeros(npts,Float)
    vc = zeros(npts,Float)
    for i in range(npts):
        rho = 0.5*dens[i] # Density of the alpha spin
        ecab,vca,vcb = cvwn(rho,rho)
        ec[i] = ecab
        vc[i] = vca
    return ec,vc

def B(dens,gamma):
    """
    Becke 1988 Exchange Functional. From 'Density-functional exchange
    energy approximation with correct asymptotic behavior.' AD Becke,
    PRA 38, 3098 (1988).
    """
    npts = len(dens)
    assert len(gamma) == npts
    ex = zeros(npts,'d')
    vx = zeros(npts,'d')
    for i in range(npts):
        rho = 0.5*dens[i]  # Density of the alpha spin
        gam = 0.25*gamma[i]
        exa,vxa = xb(rho,gam)
        ex[i] = 2*exa
        vx[i] = vxa
    return ex,vx

def LYP(dens,gamma):
    """Transformed version of LYP. See 'Results obtained with correlation
    energy density functionals of Becke and Lee, Yang, and Parr.' Miehlich,
    Savin, Stoll and Preuss. CPL 157, 200 (1989).
    """
    npts = len(dens)
    ec = zeros(npts,Float)
    vc = zeros(npts,Float)
    for i in range(npts):
        rho = 0.5*dens[i] # Density of the alpha spin
        gam = 0.25*gamma[i]
        ecab,vca,vcb = clyp(rho,rho,gam,gam,gam)
        ec[i] = ecab
        vc[i] = vca
    return ec,vc

def XPBE(dens,gamma):
    "PBE Exchange Functional"
    npts = len(dens)
    assert len(gamma) == npts
    ex = zeros(npts,'d')
    vx = zeros(npts,'d')
    for i in range(npts):
        rho = 0.5*dens[i]  # Density of the alpha spin
        gam = 0.25*gamma[i]
        exa,vxa = xpbe(rho,gam)
        ex[i] = 2*exa
        vx[i] = vxa
    return ex,vx

def CPBE(dens,gamma):
    "PBE Correlation Functional"
    npts = len(dens)
    ec = zeros(npts,Float)
    vc = zeros(npts,Float)
    for i in range(npts):
        rho = 0.5*dens[i] # Density of the alpha spin
        gam = 0.25*gamma[i]
        ecab,vca,vcb = cpbe(rho,rho,gam,gam,gam)
        ec[i] = ecab
        vc[i] = vca
    return ec,vc

# Functional terms themselves.
# Functionals are defined in their spin-polarized versions. However,
# since the exchange functionals factor (e.g. ex(rhoa,rhob) = ex(rhoa)+ex(rhob))
# we define them in the separate form, understanding that we can always
# use the above relationship to refactor them
def xs(rho,**opts):
    "Xalpha X functional. Can pass in 'alpha' as a kwarg"
    tol = opts.get('tol',1e-10)
    Xalpha = opts.get('Xalpha',2./3.)
    fac=-2.25*Xalpha*pow(3./4./pi,1./3.)
    if rho < tol: rho=0
    rho3 = pow(rho,1./3.)
    ex = fac*rho*rho3
    vx = (4./3.)*fac*rho3
    return ex,vx

def xb(rho,gam,**opts):
    tol = opts.get('tol',1e-10)
    return_flag = opts.get('return_flag',0)
    ex=vxrho=vxgam=0
    if rho > tol:
        rho13 = pow(rho,1./3.)
        x = sqrt(gam)/rho13/rho
        g = b88_g(x)
        dg = b88_dg(x)
        vxrho = (4./3.)*rho13*(g-x*dg)
        if gam > tol: vxgam = 0.5*dg/sqrt(gam)
        ex = rho*rho13*g
    if return_flag == 1: return ex,vxrho,vxgam
    return ex,vxrho

def xpbe(rho,gam,**opts):
    tol = opts.get('tol',1e-10)
    return_flag = opts.get('return_flag',0)
    kap = 0.804
    mu = 0.449276922095889E-2
    ex=vxrho=vxgam=0
    if rho > tol:
        ex0,vx0 = xs(rho)
        rho13 = rho**(1.E0/3.E0)
        rho43 = rho13*rho
        den = 1.E0+mu*gam/rho43/rho43
        F = 1+kap-kap/den
        ex = ex0*F
        dFdr = -(8./3.)*kap*mu*gam/den/den*rho**(-11./3.)
        vxrho = vx0*F+ex0*dFdr
        dFdg = -kap*mu/rho43/rho43/den/den
        vxgam = ex0*dFdg
    if return_flag == 1: return ex,vxrho,vxgam
    return ex,vxrho


def cvwn(rhoa,rhob,**opts):
    tol = opts.get('tol',1e-10)
    rho = rhoa+rhob
    ec = vcrhoa = vcrhob = 0
    if rho < tol: return ec,vcrhoa,vcrhob
    zeta=(rhoa-rhob)/rho
    x = pow(3./4./pi/rho,1./6.)
    epsp = vwn_epsp(x)
    depsp = vwn_depsp(x)
    g = vwn_g(zeta)
    # Can uncomment for a little more speed:
    if g < tol: return epsp*rho,epsp-(x/6.)*depsp,epsp-(x/6.)*depsp
    epsf = vwn_epsf(x)
    eps = epsp + g*(epsf-epsp)
    ec = eps*rho
    depsf = vwn_depsf(x)
    dg = vwn_dg(zeta)
    deps_dx = depsp + g*(depsf-depsp)
    deps_dg = (epsf-epsp)*dg
    vcrhoa = eps - (x/6.)*deps_dx + deps_dg*(1-zeta)
    vcrhob = eps - (x/6.)*deps_dx - deps_dg*(1+zeta)
    return ec,vcrhoa,vcrhob

def clyp(rhoa,rhob,gama,gamb,gamab,**opts):
    tol = opts.get('tol',1e-10)
    return_flag = opts.get('return_flag',0)
    a = 0.04918  # Parameters from the LYP papers
    b = 0.132
    c = 0.2533
    d = 0.349
    rho = rhoa+rhob
    ec=vcrhoa=vcrhob=vcgama=vcgamb=vcgamab=0
    if rho > tol:
        rhom3 = pow(rho,-1./3.)
        w = exp(-c*rhom3)/(1+d*rhom3)*pow(rho,-11./3.)
        dl = c*rhom3+d*rhom3/(1+d*rhom3)

        vcgama = -a*b*w*((1./9.)*rhoa*rhob*(1-3*dl-(dl-11)*rhoa/rho)-rhob*rhob)
        vcgamb = -a*b*w*((1./9.)*rhoa*rhob*(1-3*dl-(dl-11)*rhob/rho)-rhoa*rhoa)
        vcgamab = -a*b*w*((1./9.)*rhoa*rhob*(47-7*dl)-(4./3.)*rho*rho)

        ec = -4*a/(1+d*rhom3)*rhoa*rhob/rho \
             -pow(2,11./3.)*0.3*pow(3*pi*pi,2./3.)*a*b*w \
             *rhoa*rhob*(pow(rhoa,8./3.)+pow(rhob,8./3.)) \
             + vcgama*gama + vcgamb*gamb + vcgamab*gamab

        dw = -(1./3.)*pow(rho,-4./3.)*w*(11*pow(rho,1./3.)-c-d/(1+d*rhom3))
        ddl = (1./3.)*(d*d*pow(rho,-5./3.)/pow(1+d*rhom3,2)-dl/rho)

        d2e_dradga = dw/w*vcgama - a*b*w*(
            (1./9.)*rhob*(1-3*dl-(dl-11)*rhoa/rho)
            -(1./9.)*rhoa*rhob*((3+rhoa/rho)*ddl+(dl-11)*rhob/rho/rho))
        d2e_dradgb = dw/w*vcgamb - a*b*w*(
            (1./9.)*rhob*(1-3*dl-(dl-11)*rhob/rho)
            -(1./9.)*rhoa*rhob*((3+rhob/rho)*ddl+(dl-11)*rhob/rho/rho)
            -2*rhoa)
        d2e_dradgab = dw/w*vcgamab-a*b*w*(
            (1./9)*rhob*(47-7*dl)-(7./9.)*rhoa*rhob*dl-(8./3.)*rho)
        d2e_drbdga = dw/w*vcgama - a*b*w*(
            (1./9.)*rhoa*(1-3*dl-(dl-11)*rhob/rho)
            -(1./9.)*rhoa*rhob*((3+rhob/rho)*ddl+(dl-11)*rhoa/rho/rho))
        d2e_drbdgb = dw/w*vcgamb - a*b*w*(
            (1./9.)*rhoa*(1-3*dl-(dl-11)*rhoa/rho)
            -(1./9.)*rhoa*rhob*((3+rhoa/rho)*ddl+(dl-11)*rhoa/rho/rho)
            -2*rhob)
        d2e_drbdgab = dw/w*vcgamab-a*b*w*(
            (1./9)*rhoa*(47-7*dl)-(7./9.)*rhoa*rhob*dl-(8./3.)*rho)
        vcrhoa = -4*a/(1+d*rhom3)*rhoa*rhob/rho*(
            (1./3.)*d*pow(rho,-4./3.)/(1+d*rhom3)+1/rhoa-1/rho)\
            -pow(2,11./3.)*0.3*pow(3*pi*pi,2./3.)*a*b*(
            dw*rhoa*rhob*(pow(rhoa,8./3.)+pow(rhob,8./3.))
            +w*rhob*((11./3.)*pow(rhoa,8./3.)+pow(rhob,8./3.))) \
            +d2e_dradga*gama + d2e_dradgb*gamb + d2e_dradgab*gamab
        vcrhob = -4*a/(1+d*rhom3)*rhoa*rhob/rho*(
            (1./3.)*d*pow(rho,-4./3.)/(1+d*rhom3)+1/rhob-1/rho)\
            -pow(2,11./3.)*0.3*pow(3*pi*pi,2./3.)*a*b*(
            dw*rhoa*rhob*(pow(rhob,8./3.)+pow(rhoa,8./3.))
            +w*rhoa*((11./3.)*pow(rhob,8./3.)+pow(rhoa,8./3.))) \
            +d2e_drbdga*gama + d2e_drbdgb*gamb + d2e_drbdgab*gamab
    #Should also return the vcgam's when I can use them:
    if return_flag == 1: return ec,vcrhoa,vcrhob,vcgama,vcgamb,vcgamab
    return ec,vcrhoa,vcrhob

def cpbe(rhoa,rhob,gama,gamb,gamab,**opts):
    # My attempt to rewrite the functional. Not finished yet.
    tol = opts.get('tol',1e-10)
    return_flag = opts.get('return_flag',0)
    rho = rhoa+rhob
    ec = vca = vcb = vcgama = vcgamb = vcgamab = 0
    gam = 0.031091
    ohm = 0.046644
    bet = 0.066725
    if rho > tol:
        Rs = pow(3./(4.*pi*rho),1./3.)
        Zeta = (rhoa-rhob)/rho
        Kf = pow(3*pi*pi*rho,1./3.)
        Ks = sqrt(4*Kf/pi)
        Phi = 0.5*(pow(1+Zeta,2./3.) + pow(1-Zeta,2./3.))
        Phi3 = Phi*Phi*Phi
        gradrho = sqrt(gama+gamb+2.*gamab)
        T = gradrho/(2*Phi*Ks*rho)
        T2 = T*T
        T4 = T2*T2

        eps,vc0a,vc0b = cpbe_lsd(rhoa,rhob)

        expo = (exp(-eps/(gam*Phi3))-1.)
        A = bet/gam/expo
        N = T2+A*T4
        D = 1.+A*T2+A*A*T4
        H = gam*Phi3*log(1.+(bet/gam)*N/D)
        ec = rho*(eps+H)

        # Derivative stuff
        dZ_drhoa = (1.-Zeta)/rho
        dZ_drhob = -(1.+Zeta)/rho

        dPhi_dZ = pow(1.+Zeta,-1./3.)/3.-pow(1.-Zeta,-1./3.)/3.
        dPhi_drhoa = dPhi_dZ*dZ_drhoa
        dPhi_drhob = dPhi_dZ*dZ_drhob
        
        dKs_drho = Ks/(6*rho)
        
        dT_dPhi = -T/Phi
        dT_dKs = -T/Ks
        dT_drhoa = -T/rho + dT_dPhi*dPhi_drhoa + dT_dKs*dKs_drho
        dT_drhob = -T/rho + dT_dPhi*dPhi_drhob + dT_dKs*dKs_drho

        dA_dPhi = -A/expo*exp(-eps/(gam*Phi3))*(3*eps/(gam*Phi3*Phi))
        dA_deps = -A/expo*exp(-eps/(gam*Phi3))*(-1/(gam*Phi3))
        deps_drhoa = (vc0a-eps)/rho
        deps_drhob = (vc0b-eps)/rho
        dA_drhoa = dA_dPhi*dPhi_drhoa + dA_deps*deps_drhoa
        dA_drhob = dA_dPhi*dPhi_drhob + dA_deps*deps_drhoa

        dN_dT = 2*T+4*A*T2*T
        dD_dT = 2*A*T + 4*A*A*T*T2
        dN_dA = T4
        dD_dA = T2+2*A*T4

        dH_dPhi = 3*H/Phi
        dH_dT = bet*Phi3/(1.+bet/gam*N/D)*(D*dN_dT-N*dD_dT)/D/D
            
        dH_dA = bet*Phi3/(1.+bet/gam*N/D)*(D*dN_dA-N*dD_dA)/D/D
        
        dH_drhoa = dH_dPhi*dPhi_drhoa + dH_dT*dT_drhoa + dH_dA*dA_drhoa
        dH_drhob = dH_dPhi*dPhi_drhob + dH_dT*dT_drhob + dH_dA*dA_drhob
        
        vca = vc0a + H + rho*dH_drhoa
        vcb = vc0b + H + rho*dH_drhob
    # Havent done the dE_dgamma derives yet
    return ec,vca,vcb

def cpbe_lsd(rhoa,rhob):
    # Not quite VWN
    # LSD terms
    rho = rhoa+rhob
    Rs = pow(3./(4.*pi*rho),1./3.)
    Zeta = (rhoa-rhob)/rho
    thrd = 1./3.     # thrd*=various multiples of 1/3
    thrd4 = 4*thrd
    ggam=0.5198420997897463295344212145565 # gam= 2^(4/3)-2
    fzz=8./(9.*ggam) # fzz=f''(0)= 8/(9*gam)
    rtrs = sqrt(Rs)
    eu,eurs = pbe_gcor(0.0310907,0.21370,7.5957,
                       3.5876,1.6382,0.49294,rtrs)
    ep,eprs = pbe_gcor(0.01554535,0.20548,14.1189,
                       6.1977,3.3662,0.62517,rtrs)
    alfm,alfrsm = pbe_gcor(0.0168869,0.11125,10.357,
                           3.6231,0.88026,0.49671,rtrs)
    alfc = -alfm
    z4 = Zeta**4
    f=((1.+Zeta)**thrd4+(1.-Zeta)**thrd4-2.)/ggam
    eps = eu*(1.-f*z4)+ep*f*z4-alfm*f*(1.-z4)/fzz

    ecrs = eurs*(1.-f*z4)+eprs*f*z4-alfrsm*f*(1.-z4)/fzz
    fz = thrd4*((1.+Zeta)**thrd-(1.-Zeta)**thrd)/ggam
    eczet = 4.*(Zeta**3)*f*(ep-eu+alfm/fzz)+fz*(z4*ep-z4*eu-(1.-z4)*alfm/fzz)
    comm = eps -Rs*ecrs/3.-Zeta*eczet
    vca = comm + eczet
    vcb = comm - eczet
    return eps,vca,vcb
    
def pbe_gcor(a,a1,b1,b2,b3,b4,rtrs):
#      subroutine gcor2(a,a1,b1,b2,b3,b4,rtrs,gg,ggrs)
# slimmed down version of gcor used in pw91 routines, to interpolate
# lsd correlation energy, as given by (10) of
# j. p. perdew and y. wang, phys. rev. b {\bf 45}, 13244 (1992).
# k. burke, may 11, 1996.
#      implicit real*8 (a-h,o-z)
      q0 = -2.*a*(1.+a1*rtrs*rtrs)
      q1 = 2.*a*rtrs*(b1+rtrs*(b2+rtrs*(b3+b4*rtrs)))
      q2 = log(1.+1./q1)
      gg = q0*q2
      q3 = a*(b1/rtrs+2.*b2+rtrs*(3.*b3+4.*b4*rtrs))
      ggrs = -2.*a*a1*q2-q0*q3/(q1*(1.+q1))
      return gg,ggrs

def vwn_xx(x,b,c): return x*x+b*x+c
def vwn_epsp(x): return vwn_eps(x,0.0310907,-0.10498,3.72744,12.9352)
def vwn_epsf(x): return vwn_eps(x,0.01554535,-0.32500,7.06042,13.0045)
def vwn_eps(x,a,x0,b,c):
    q = sqrt(4*c-b*b)
    eps = a*(log(x*x/vwn_xx(x,b,c))
             - b*(x0/vwn_xx(x0,b,c))*log(pow(x-x0,2)/vwn_xx(x,b,c))
             + (2*b/q)*(1-(x0*(2*x0+b)/vwn_xx(x0,b,c))) * atan(q/(2*x+b)))
    return eps

def vwn_depsp(x): return vwn_deps(x,0.0310907,-0.10498,3.72744,12.9352)
def vwn_depsf(x): return vwn_deps(x,0.01554535,-0.32500,7.06042,13.0045)
def vwn_deps(x,a,x0,b,c):
    q = sqrt(4*c-b*b)
    deps = a*(2/x - (2*x+b)/vwn_xx(x,b,c)
              - 4*b/(pow(2*x+b,2)+q*q) - (b*x0/vwn_xx(x0,b,c))
              * (2/(x-x0)-(2*x+b)/vwn_xx(x,b,c)-4*(2*x0+b)/(pow(2*x+b,2)+q*q)))
    return deps

def vwn_g(z): return 1.125*(pow(1+z,4./3.)+pow(1-z,4./3.)-2)
def vwn_dg(z): return 1.5*(pow(1+z,1./3.)-pow(1-z,1./3.))

def b88_g(x,b=0.0042):
    return -1.5*pow(3./4./pi,1./3.)-b*x*x/(1.+6.*b*x*arcsinh(x))

def b88_dg(x,b=0.0042):
    num = 6*b*b*x*x*(x/sqrt(x*x+1)-arcsinh(x))-2*b*x
    denom = pow(1+6*b*x*arcsinh(x),2)
    return num/denom

def pbe_F(s,mu=0.2195149727645171,kap=0.8040):
    P0 = 1.+mu*s*s/kap
    return 1.+kap-kap/P0

def pbe_dF(s,mu=0.2195149727645171,kap=0.8040):
    P0 = 1.+mu*s*s/kap
    return 2*mu*s/(P0*P0)


# Table mapping functional names into functions etc.
# LDA -> SVWN, GGA -> PBE
xfuncs = dict(LDA=S,S0=S,SVWN=S,BLYP=B,PBE=XPBE)
cfuncs = dict(LDA=None,S0=None,SVWN=VWN,BLYP=LYP,PBE=CPBE)
need_gradients = dict(LDA=False,S0=False,SVWN=False,BLYP=True,PBE=True)


