#!/usr/bin/env python
"""\
 fermi_dirac.py: Utilities for finite temperature Fermi-Dirac occupations.

 This program is part of the PyQuante quantum chemistry program suite.

 Copyright (c) 2004, Richard P. Muller. All Rights Reserved. 

 PyQuante version 1.2 and later is covered by the modified BSD
 license. Please see the file LICENSE that is part of this
 distribution. 
"""

#  Tested on NO. Seems to work pretty well, although (not surprisingly)
#   it gives a result slightly different from the UHF result. Appears to
#   require averaging, which makes the code converge very slowly. Both zero
#   temperature and finite temperature appear to work

import sys
from Numeric import sum,matrixmultiply,transpose
from math import exp
from Constants import Kboltz
from LA2 import mkdens

def mkdens_occs(c,occs,**opts):
    """Density matrix from a set of occupations (e.g. from FD expression).
    """
    tol = opts.get('tol',1e-5)
    verbose = opts.get('verbose',False)
    # Determine how many orbs have occupations greater than 0
    norb = 0
    for fi in occs:
        if fi < tol: break
        norb += 1
    if verbose:
        print "mkdens_occs: %d occupied orbitals found" % norb
    # Determine how many doubly occupied orbitals we have
    nclosed = 0
    for i in range(norb):
        if abs(1.-occs[i]) > tol: break
        nclosed += 1
    if verbose:
        print "mkdens_occs: %d closed-shell orbitals found" % nclosed
    D = mkdens(c,0,nclosed)
    for i in range(nclosed,norb):
        d = c[i:i+1,:]
        D = D + occs[i]*matrixmultiply(transpose(d),d)
    return D
    
def get_fermi_occ(efermi,en,temp):
    kT = Kboltz*temp
    x = (en-efermi)/kT
    if x < -50.: return 1.
    elif x > 50.: return 0
    return 1/(1+exp(x))
    

def get_fermi_occs(efermi,orbe,temp):
    occs = []
    for en in orbe:
        occs.append(get_fermi_occ(efermi,en,temp))
    return occs

def get_t0_occs(nel,nbf):
    occs = [0]*nbf
    nc,no = divmod(nel,2)
    for i in range(nc): occs[i] = 1.
    for i in range(nc,nc+no): occs[i] = 0.5
    return occs

def get_efermi(nel,orbe,temp,**opts):
    "Bisection method to get Fermi energy from Fermi-Dirac dist"
    tol = opts.get('tol',1e-9)
    verbose = opts.get('verbose',False)

    if verbose:
        print "Attempting to compute E Fermi"
        print "OrbE = ",orbe
    elow,ehigh = orbe[0],orbe[-1]
    nlow = 2*sum(get_fermi_occs(elow,orbe,temp))
    nhigh = 2*sum(get_fermi_occs(ehigh,orbe,temp))

    assert nlow < nel, "elow incorrect %f " % nlow
    assert nhigh > nel, "ehigh incorrect %f" % nhigh
    for i in range(100):
        efermi = (elow+ehigh)/2
        n = 2*sum(get_fermi_occs(efermi,orbe,temp))
        if abs(n-nel) < tol:
            break
        elif n < nel:
            elow = efermi
        else:
            ehigh = efermi
    else:
        print "get_fd_occs: Too many iterations"
    return efermi

# Commented out 11/18/2005 by RPM
# Remove if no one notices for 2 months.
# def fdscf(atoms,S,h,Ints,charge=0,ConvCriteria=0.0001,MaxIter=20):
#     orbe,orbs = GHeigenvectors(h,S)
#     nel = atoms.get_nel(charge)
#     nbf = S.shape[0]
#     efermi = orbe[int(nel/2)]
#     enuke = atoms.get_enuke()
#     eold = 0.
#     for i in range(MaxIter):
#         efermi = get_efermi(nel,orbe,efermi)
#         occs = get_fermi_occs(efermi,orbe)
#         D = get_dens(orbs,occs)
#         if averaging:
#             if i: # do averaging
#                 D = averaging*D + (1-averaging)*D0
#             D0 = D
        
#         G = get2JmK(Ints,D)
#         F = h+G
#         orbe,orbs = GHeigenvectors(F,S)
#         energy = get_energy(h,F,D,enuke)
#         #print i,energy
#         if abs(energy-eold) < ConvCriteria: break
#         eold = energy
#     return energy

# def fdhf(atoms):
#     bfs = getbasis(atoms)
#     S,h = get1ints(bfs,atoms)
#     Ints = get2ints(bfs)
#     energy = fdscf(atoms,S,h,Ints)
#     return energy
